from typing import Tuple
import numpy as np
from .sndfileio import BACKENDS


def compare_read(path:str, backend) -> Tuple[np.ndarray, np.ndarray]:
    backends = [backend for backend in BACKENDS if backend.is_available()]
    backend1 = next(b for b in backends if b.name == backend)
    backend2 = next(b for b in backends if b.name == 'builtin')
    s0 = backend1.read(path)
    s1 = backend2.read(path)
    return s0, s1


def testfile(path:str, exact=True) -> bool:
    s0, s1 = compare_read(path)
    if exact:
        return (s0[0] == s1[0]).all()
    return np.allclose(s0, s1)


"""

This package providesa unified API to read and write sound-files to and from
numpy arrays. It can read and write to `wav`, `aif`, `flac`, `mp3` and
all loss-less formats supported by `libsndfile`.

--------------------------------------------------

API
****

Read / write a file in one function.

:func:`sndinfo<sndfileio.sndinfo>`
==================================

    Returns a :class:`SndInfo` with all the information of the sound-file

:func:`sndread<sndfileio.sndread>`
==================================

    Reads ALL the samples. Returns a tuple (data, samplerate)

:func:`sndwrite<sndfileio.sndwrite>`
====================================

    Write samples to outfile

:func:`sndwrite_like<sndfileio.sndwrite_like>`
==============================================

    Write samples to outfile cloning another files format & encoding


:func:`sndget<sndfileio.sndget>`
==================================

    Reads the sample data, returns a tuple (**samples**, **info**), where
    `info` is a :class:`SndInfo` with all the information of the sound-file

-----------------------------


==========
Chunked IO
==========

It is possible to stream a soundfile by reading and processing chunks. This
is helpful in order to avoid allocating memory for a large. The same is possible
for writing

:func:`sndread_chunked<sndfileio.sndread_chunked>`
==================================================

    Returns a generator yielding chunks of frames

:func:`sndwrite_chunked<sndfileio.sndwrite_chunked>`
====================================================

    Opens the file for writing. Returns a :class:`SndWriter`. To write to the file,
    call :meth:`write` on the returned handle


---------


Examples
--------

.. code-block:: python

    # Normalize and save as flac
    from sndfileio import sndread, sndwrite
    samples, sr = sndread("in.wav")
    maxvalue = max(samples.max(), -samples.min())
    samples *= 1/maxvalue
    sndwrite(samples, sr, "out.flac")


.. code-block:: python

    # Process a file in chunks
    from sndfileio import *
    from sndfileio.dsp import
    with sndwrite_chunked(44100, "out.flac") as writer:
        for buf in sndread_chunked("in.flac"):
            # do some processing, like changing the gain
            buf *= 0.5
            writer.write(buf)


"""

from .sndfileio import *
from .resampling import resample
from . import util

del resampling



from __future__ import annotations
import scipy.signal as sig
import numpy as np
from .dsp import lowpass_cheby
import logging
from math import gcd
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import List, Optional as Opt, Callable


class BackendNotAvailable(Exception):
    pass


logger = logging.getLogger("sndfileio")


def _applyMultichan(samples: np.ndarray,
                    func: Callable[[np.ndarray], np.ndarray]) -> np.ndarray:
    """
    Apply func to each channel of audio data in samples
    """
    if len(samples.shape) == 1 or samples.shape[1] == 1:
        newsamples = func(samples)
    else:
        y = np.array([])
        for i in range(samples.shape[1]):
            y = np.concatenate((y, func(samples[:,i])))
        newsamples = y.reshape(samples.shape[1], -1).T
    return newsamples    
    

def _resample_scipy(samples: np.ndarray, sr:int, newsr:int, window='hanning'
                    ) -> np.ndarray:
    try:
        from scipy.signal import resample
    except ImportError:
        raise BackendNotAvailable()

    ratio = newsr/sr
    lenNewSamples = int(ratio * len(samples) + 0.5)

    return _applyMultichan(samples, 
                           lambda S: resample(S, lenNewSamples, window=window))


def _resample_samplerate(samples:np.ndarray, sr:int, newsr:int) -> np.ndarray:
    """
    Uses https://github.com/tuxu/python-samplerate
    """
    try:
        from samplerate import resample
    except ImportError:
        raise BackendNotAvailable()

    ratio = newsr/sr
    return _applyMultichan(samples,
                           lambda S: resample(S, ratio, 'sinc_best'))

#######################################################

# global cache of resamplers
_precomputed_filters = {}


def _nnresample_compute_filt(up, down, beta=5.0, L=32001):
    r"""
    Computes a filter to resample a signal from rate "down" to rate "up"
    
    Parameters
    ----------
    up : int
        The upsampling factor.
    down : int
        The downsampling factor.
    beta : float
        Beta factor for Kaiser window.  Determines tradeoff between
        stopband attenuation and transition band width
    L : int
        FIR filter order.  Determines stopband attenuation.  The higher
        the better, ath the cost of complexity.
        
    Returns
    -------
    filt : array
        The FIR filter coefficients
        
    Notes
    -----
    This function is to be used if you want to manage your own filters
    to be used with scipy.signal.resample_poly (use the `window=...`
    parameter).  WARNING: Some versions (at least 0.19.1) of scipy
    modify the passed filter, so make sure to make a copy beforehand:
    
    out = scipy.signal.resample_poly(in up, down, window=numpy.array(filt))
    """
    
    # Determine our up and down factors
    g = gcd(up, down)
    up = up//g
    down = down//g
    max_rate = max(up, down)

    sfact = np.sqrt(1+(beta/np.pi)**2)
            
    # generate first filter attempt: with 6dB attenuation at f_c.
    init_filt = sig.fir_filter_design.firwin(L, 1/max_rate, window=('kaiser', beta))
    
    # convert into frequency domain
    N_FFT = 2**19
    NBINS = N_FFT/2+1
    paddedfilt = np.zeros(N_FFT)
    paddedfilt[:L] = init_filt
    ffilt = np.fft.rfft(paddedfilt)
    
    # now find the minimum between f_c and f_c+sqrt(1+(beta/pi)^2)/L
    bot = int(np.floor(NBINS/max_rate))
    top = int(np.ceil(NBINS*(1/max_rate + 2*sfact/L)))
    firstnull = (np.argmin(np.abs(ffilt[bot:top])) + bot)/NBINS
    
    # generate the proper shifted filter
    return sig.fir_filter_design.firwin(L, -firstnull+2/max_rate, window=('kaiser', beta))


def _resample_nnresample(samples: np.ndarray, sr:int, newsr:int) -> np.ndarray:
    return _applyMultichan(samples,
                           lambda S: _resample_nnresample2(S, newsr, sr)[:-1])


def _resample_nnresample_package(samples: np.ndarray, sr:int, newsr:int) -> np.ndarray:
    return _applyMultichan(samples,
                           lambda S: _resample_nnresample_package_mono(S, newsr, sr)[:-1])


def _resample_nnresample_package_mono(s:np.ndarray, up:int, down:int, **kws) -> np.ndarray:
    import nnresample
    return nnresample.resample(s, up, down, axis=0, fc='nn', **kws)


def _resample_nnresample2(s:np.ndarray, up:int, down:int, beta=5.0, L=16001, axis=0
                          ) -> np.ndarray:
    """
    Taken from https://github.com/jthiem/nnresample

    Resample a signal from rate "down" to rate "up"

    Args:
        s (array): The data to be resampled.
        up (int): The upsampling factor.
        down (int): The downsampling factor.
        beta (float): Beta factor for Kaiser window. Determines tradeoff between
            stopband attenuation and transition band width
        L (int): FIR filter order.  Determines stopband attenuation.  The higher
            the better, ath the cost of complexity.
        axis (int): int, optional. The axis of `s` that is resampled. Default is 0.
        
    Returns:
        The resampled array.

    .. note::
    
        The function keeps a global cache of filters, since they are
        determined entirely by up, down, beta, and L.  If a filter
        has previously been used it is looked up instead of being
        recomputed.
    """
    # check if a resampling filter with the chosen parameters already exists
    params = (up, down, beta, L)
    if params in _precomputed_filters.keys():
        # if so, use it.
        filt = _precomputed_filters[params]
    else:
        # if not, generate filter, store it, use it
        filt = _nnresample_compute_filt(up, down, beta, L)
        _precomputed_filters[params] = filt
    return sig.resample_poly(s, up, down, window=np.array(filt), axis=axis)


def _resample_obspy(samples:np.ndarray, sr:int, newsr:int, window='hanning', lowpass=True
                    ) -> np.ndarray:
    """
    Resample using Fourier method. The same as resample_scipy but with
    low-pass filtering for upsampling
    """
    from scipy.signal import resample
    from math import ceil
    factor = sr/float(newsr)
    if newsr < sr and lowpass:
        # be sure filter still behaves good
        if factor > 16:
            logger.info("Automatic filter design is unstable for resampling "
                        "factors (current sampling rate/new sampling rate) " 
                        "above 16. Manual resampling is necessary.")
        freq = min(sr, newsr) * 0.5 / float(factor)
        logger.debug(f"resample_obspy: lowpass {freq}")
        samples = lowpass_cheby(samples, freq=freq, sr=sr, maxorder=12)
    num = int(ceil(len(samples) / factor))

    return _applyMultichan(samples, 
                           lambda S: resample(S, num, window=window))


def resample(samples: np.ndarray, oldsr:int, newsr:int) -> np.ndarray:
    """
    Resample `samples` with given samplerate `sr` to new samplerate `newsr`

    Args:
        samples: mono or multichannel frames
        oldsr: original samplerate
        newsr: new sample rate

    Returns:
        the new samples
    """
    backends = [
        _resample_samplerate,   # turns the samples into float32, which is ok for audio
        _resample_nnresample_package,  # nnresample packaged version
        _resample_nnresample,   # (builtin) very good results, follows libsamplerate closely
        _resample_obspy,        # these last two introduce some error at the first samples
        _resample_scipy
    ]

    for backend in backends:
        try:
            return backend(samples, oldsr, newsr)
        except BackendNotAvailable:
            pass



from __future__ import annotations
import numpy as np
import os


from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Callable, Optional as Opt, Iterator, Tuple, Union



encodings_for_format = {
    'wav': ['pcm16', 'pcm24', 'pcm32', 'float32', 'float64'],
    'aiff': ['pcm16', 'pcm24', 'pcm32', 'float32', 'float64'],
    'flac': ['pcm16', 'pcm24'],
    'mp3': ['pcm16', 'pcm24'],
    'ogg': ['pcm16']
}

_default_encoding = {
    'wav': 'float32',
    'aif': 'float32',
    'aiff': 'float32',
    'mp3': 'pcm16',
    'flac': 'pcm24',
    'ogg': 'pcm16'
}


metadata_possible_keys = {'comment', 'title', 'artist', 'album', 'tracknumber', 'software'}


known_fileformats = encodings_for_format.keys()


def getchannel(samples: np.ndarray, channel:int) -> np.ndarray:
    """
    Returns a view into a channel of samples.

    Args:
        samples: a numpy array representing the audio data
        channel: the channel to extract (channels begin with 0)

    Returns:
        the channel specified, as a numpy array

    Example
    ~~~~~~~
    ::

        # Read a stereo file, atenuate one channel
        >>> from sndfileio import *
        >>> stereosamples, sr = sndread("stereo.wav")
        >>> ch0 = util.getchannel(stereosamples, 0)
        >>> ch0 *= 0.5
        >>> sndwrite_like("out.wav", stereosamples, likefile="stereo.wav")

    """
    N = numchannels(samples)
    if channel > (N - 1):
        raise ValueError(f"channel {channel} out of range (max. {N} channels)")
    if N == 1:
        return samples
    return samples[:, channel]


def numchannels(samples:np.ndarray) -> int:
    """
    return the number of channels present in samples

    Args:
        samples: a numpy array as returned by sndread

    Returns:
        the number of channels in `samples`

    For multichannel audio, samples is always interleaved,
    meaning that samples[n] returns always a frame, which
    is either a single scalar for mono audio, or an array
    for multichannel audio.
    """
    if len(samples.shape) == 1:
        return 1
    else:
        return samples.shape[1]


def apply_multichannel(data:np.ndarray, func:Callable[[np.ndarray], np.ndarray]
                       ) -> np.ndarray:
    """
    Apply a function ``(samples1D) -> samples1D`` along the channels of a multichannel
    sample array, returning a multichannel sample array of the same shape

    Args:
        data: the samples
        func: the function to apply to each channel

    Returns:
        the resulting samples

    Example::

        >>> import numpy as np
        >>> source = np.array([[0.,  0.],
        ...                    [0.1, 0.2],
        ...                    [0.3, 0.4.],
        ...                    [0.5, 0.6]])
        >>> apply_multichannel(source, lambda chan: chan*0.5)
        array([[0.,   0. ]
               [0.05, 0.1]
               [0.15, 0.2]
               [0.25, 0.3]])
    """
    ch = numchannels(data)
    if ch == 1:
        return func(data)
    else:
        chans = []
        for i in range(ch):
            chans.append(func(data[:,i]))
        out = np.concatenate(chans)
        out.shape = (ch, len(data))
        return out.T


def numpy24to32bit(data:np.ndarray, bigendian:bool=False) -> np.ndarray:
    """
    Convert a 24-bit pcm to 32-bit pcm samples

    Args:
        data: a ubyte array of shape = (size,) (interleaved channels if multichannel)
        bigendian: is the data represented as big endian?
    """
    target_size = int(data.shape[0] * 4 / 3.)
    target = np.zeros((target_size,), dtype=np.ubyte)
    if not bigendian:
        target[3::4] = data[2::3]
        target[2::4] = data[1::3]
        target[1::4] = data[0::3]
    else:
        target[1::4] = data[2::3]
        target[2::4] = data[1::3]
        target[3::4] = data[0::3]
    targetraw = target.tobytes()
    del target
    return np.fromstring(targetraw, dtype=np.int32)


def mix(samples:np.ndarray, scale_by_numchannels:bool=True) -> np.ndarray:
    summed = samples.sum(0)
    if scale_by_numchannels:
        summed *= (1 / numchannels(samples))
    return summed


def fileformat_from_ext(ext: str) -> Opt[str]:
    """
    Deduces the file format from the extension

    The returned file format is always a lowercase string,
    one of "wav", "aiff", "flac", "ogg", "wv", "mp3".
    Reurns None if not a supported extension

    Examples
    --------

    >>> fileformat_from_ext(("out.wav"))
    wav
    >>> fileformat_from_ext(("foo.FLAC"))
    flac
    >>> fileformat_from_ext(("bar.aif"))
    aiff

    """
    if ext.startswith("."):
        ext = ext[1:]
    ext = ext.lower()
    if ext == "aif":
        return "aiff"
    return ext if ext in known_fileformats else None


def detect_format(path:str) -> Opt[str]:
    """
    Detect the file format of a given soundfile

    Args:
        path: the path to the soundfile

    Returns:
        the fileformat, or None if not a known file format
    """
    ext = os.path.splitext(path)[1][1:].lower() if path else None
    if ext in known_fileformats:
        return ext
    import filetype
    ext = filetype.guess_extension(path)
    return ext if ext in known_fileformats else None


def default_encoding(fileformat:str) -> Opt[str]:
    """
    Return the default encoding for the given fileformat

    =======    =================
    Format     Default encoding
    =======    =================
     wav        float32
     aif        float32
     flac       pcm24
     mp3        pcm16
    =======    =================

    """
    return _default_encoding.get(fileformat)


def asmono(samples:np.ndarray, channel:Union[int, str]=0) -> np.ndarray:
    """
    convert samples to mono if they are not mono already.

    The returned array will always have the shape (numframes,)

    Args:
        samples: the multichannel samples
        channel: the channel number to use, or 'mix' to mix-down all channels

    Returns:
        the samples as one mono channel
    """
    if numchannels(samples) == 1:
        return samples

    if isinstance(channel, int):
        return samples[:, channel]
    elif channel == 'mix':
        return mix(samples, scale_by_numchannels=True)
    else:
        raise ValueError("channel has to be an integer indicating a channel,"
                         " or 'mix' to mix down all channels")


def bitdepth(data:np.ndarray, snap:bool=True) -> int:
    """
    returns the number of bits actually used to represent the data.

    Args:
        data: a numpy.array (mono or multi-channel)
        snap: snap to 8, 16, 24 or 32 bits.

    Returns:
        the bits needed to represent the data
    """
    data = asmono(data)
    maxitems = min(4096, data.shape[0])
    maxbits = max(x.as_integer_ratio()[1]
                  for x in data[:maxitems]).bit_length()
    if snap:
        if maxbits <= 8:
            maxbits = 8
        elif maxbits <= 16:
            maxbits = 16
        elif maxbits <= 24:
            maxbits = 24
        elif maxbits <= 32:
            maxbits = 32
        else:
            maxbits = 64
    return maxbits


def as_float_array(data:np.ndarray, encoding:str) -> np.ndarray:
    """
    Convert (if necessary) an array containing pcm (integer) samples
    to float64 between -1:1

    Args:
        data: the samples to convert
        encoding: the encoding of data (one of 'float32', 'pcm24', 'pcm16')

    Returns:
        data represented as a float numpy array
    """
    assert (data > 0).any()
    if encoding == 'float32':
        return data
    elif encoding == 'pcm24':
        return data / (2.0 ** 31)
    elif encoding == 'pcm16':
        return data / (2.0 ** 15)
    else:
        raise ValueError("encoding not understood")


def samples_out_of_range(data:np.ndarray) -> bool:
    """
    Returns True if any sample is out of the range -1:1
    """
    import numpyx
    for i in range(numchannels(data)):
        ch = getchannel(data, i)
        if numpyx.any_less_than(ch, -1) or numpyx.any_greater_than(ch, 1):
            return True
    return False


def guess_encoding(data:np.ndarray, fmt: str) -> str:
    """
    Guess the encoding for data based on the format. For each format
    an encoding is selected which is able to represent the data without
    loss. In the case of wav/aiff, if there is sample data outside of the
    -1:1 range, a float32 encoding is chosen, since any pcm representation
    would result in out of range samples.

    Args:
        data: the samples
        fmt: on of 'wav', 'aiff', 'flac'

    Returns:
        the encoding, one of 'pcm16', 'pcm24', 'float32'
    """
    if fmt in ('wav', 'aif', 'aiff'):
        if samples_out_of_range(data):
            return 'float32'
        maxbits = min(32, bitdepth(data, snap=True))
        encoding = {
            16: 'pcm16',
            24: 'pcm24',
            32: 'flpat32',
        }.get(maxbits, 'float32')
    elif fmt == "flac":
        maxbits = min(24, bitdepth(data, snap=True))
        encoding = {
            16: 'pcm16',
            24: 'pcm24',
        }.get(maxbits, 'pcm24')
    elif fmt == 'mp3':
        encoding = 'pcm16'
    else:
        raise ValueError(f"The format {fmt} is not supported")
    assert encoding in ('pcm16', 'pcm24', 'float32')
    return encoding


def chunks(start:int, end:int, step:int) -> Iterator[Tuple[int, int]]:
    pos = start
    last_full = end - step
    while pos < last_full:
        yield pos, step
        pos += step
    yield pos, end - pos


def tinytagMetadata(path: str) -> dict:
    import tinytag
    m = tinytag.TinyTag.get(path)
    metadata = {}
    if m.title:
        metadata['title'] = m.title
    if m.album:
        metadata['album'] = m.album
    if m.comment:
        metadata['comment'] = m.comment
    if m.artist:
        metadata['artist'] = m.artist
    if m.track:
        metadata['tracknumber'] = m.track
    if m.bitrate:
        metadata['bitrate'] = m.bitrate
    return metadata



from __future__ import annotations
from .datastructs import SndInfo
from . import util
import numpy as np 
import miniaudio
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .datastructs import sample_t
    from typing import Iterator


def _readfragment(path: str, start: float, end: float
                  ) -> sample_t:
    info = miniaudio.mp3_get_file_info(path)
    seek_frame = int(start * info.sample_rate)
    if end == 0:
        frames_to_read = info.num_frames - seek_frame
    else:
        frames_to_read = int(end * info.sample_rate) - seek_frame
    buf = next(miniaudio.mp3_stream_file(path, frames_to_read=frames_to_read, seek_frame=seek_frame))
    samples = np.asarray(buf, dtype=float)
    samples /= 2**15
    return samples, info.sample_rate


def mp3read_chunked(path: str, chunksize: int, start=0., stop=0.
                    ) -> Iterator[np.ndarray]:
    info = miniaudio.mp3_get_file_info(path)
    sr = info.sample_rate
    seek_frame = int(start*sr)
    if stop == 0:
        frames_to_read = info.num_frames - seek_frame
    else:
        frames_to_read = min(info.num_frames, int(sr * stop)) - seek_frame
    assert frames_to_read > 0
    nchannels = info.nchannels
    for buf in miniaudio.mp3_stream_file(path, frames_to_read=chunksize, seek_frame=seek_frame):
        samples = np.asarray(buf, dtype=float)
        samples /= 2**15
        if nchannels > 1:
            samples.shape = (len(samples) // nchannels, nchannels)
        if frames_to_read < len(samples):
            samples = samples[:frames_to_read]
            yield samples
            return
        else:
            yield samples
            frames_to_read -= len(samples)


def mp3read(path: str, start=0., end=0.) -> sample_t:
    """
    Reads a mp3 files completely into an array
    """
    if start > 0 or end > 0:
        return _readfragment(path, start, end)
    decoded = miniaudio.mp3_read_file_f32(path)
    npsamples = np.frombuffer(decoded.samples, dtype='float32').astype(float)
    if decoded.nchannels > 1:
        npsamples.shape = (decoded.num_frames, decoded.nchannels)
    return npsamples, decoded.sample_rate


_encodingFromFormat = {
    'SIGNED16': 'pcm16'
}


def makeSndInfo(info: miniaudio.SoundFileInfo, metadata: dict = None) -> SndInfo:
    encoding = _encodingFromFormat.get(info.file_format.name, '')
    bitrate = metadata.pop('bitrate', None)
    return SndInfo(samplerate=info.sample_rate,
                   nframes=info.num_frames,
                   channels=info.nchannels,
                   encoding=encoding,
                   fileformat=info.file_format.name,
                   metadata=metadata,
                   bitrate=bitrate)


def mp3info(path: str) -> SndInfo:
    info = miniaudio.mp3_get_file_info(path)
    metadata = util.tinytagMetadata(path)
    return makeSndInfo(info, metadata=metadata)


def ogginfo(path: str) -> SndInfo:
    info = miniaudio.vorbis_get_file_info(path)
    metadata = util.tinytagMetadata(path)
    return makeSndInfo(info, metadata=metadata)


from __future__ import print_function
from setuptools import setup
import os

thisdir = os.path.abspath(os.path.dirname(__file__))
long_description = open(os.path.join(thisdir, 'README.rst')).read()

setup(
    version="1.9.1",
    name="sndfileio",
    python_requires=">=3.8",
    install_requires=[
        "numpy>1.8",
        "scipy",
        "soundfile",
        "miniaudio",
        "nnresample",
        "numpyx>=0.4.1",
        "tinytag",
        "filetype",
        "lameenc"],
    packages=["sndfileio"],
    package_data={'': ['README.rst'], 'sndfileio': ['py.typed']},
    include_package_data=True,

    # Metadata
    description = "Simple API for reading / writing soundfiles",
    long_description = long_description,
    summary      = "Simple API for reading and writing soundfiles", 
    author       = "Eduardo Moguillansky",
    author_email = "eduardo.moguillansky@gmail.com",
    url          = "https://github.com/gesellkammer/sndfileio",
    license = 'LGPL',
)



from __future__ import annotations
import os as _os
import sys as _sys
import numpy as np
from importlib.util import find_spec as _find_spec
import logging

try:
    import soundfile as _soundfile
except (IOError, ImportError) as e:
    if 'sphinx' in _sys.modules:
        from sphinx.ext.autodoc.mock import _MockObject
        _soundfile = _MockObject()
    else:
        raise e

from . import util
from .datastructs import SndInfo, SndWriter

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import pysndfile
    from .datastructs import sample_t
    from typing import Dict, Tuple, Union, Iterator, Optional as Opt, List, Callable, Type


logger = logging.getLogger("sndfileio")


__all__ = [
    "sndread",
    "sndread_chunked",
    "sndget",
    "sndinfo",
    "sndwrite",
    "sndwrite_like",
    "sndwrite_chunked",
    "sndwrite_chunked_like",
    "mp3write",
    "SndInfo",
    "SndWriter"
]


class FormatNotSupported(Exception):
    pass


def _is_package_installed(pkg):
    return _find_spec(pkg) is not None


class SndfileError(IOError):
    pass


def _normalize_path(path: str) -> str:
    return _os.path.expanduser(path)


def sndread(path: str, start: float = 0, end: float = 0) -> sample_t:
    """
    Read a soundfile as a numpy array.

    This is a float array defined between -1 and 1, independently of the format
    of the soundfile

    Args:
        path: (str) the path to read
        start: (float) the time to start reading
        end: (float) the time to end reading (0=until the end)

    Returns:
        a tuple (samples:ndarray[dtype=float], sr: int)

    Example
    ~~~~~~~

    ::

        # Normalize and save as flac
        from sndfileio import sndread, sndwrite
        samples, sr = sndread("in.wav")
        maxvalue = max(samples.max(), -samples.min())
        samples *= 1/maxvalue
        sndwrite(samples, sr, "out.flac")
    """
    path = _normalize_path(path)
    if not _os.path.exists(path):
        raise IOError(f"File not found: {path}")
    backend = _get_backend(path)
    if not backend:
        raise RuntimeError(f"No backend available to read {path}")
    logger.debug(f"sndread: using backend {backend.name}")
    return backend.read(path, start=start, end=end)


def sndread_chunked(path: str, chunksize: int = 2048, start: float = 0., stop: float = 0.
                    ) -> Iterator[np.ndarray]:
    """
    Read a soundfile in chunks

    Args:
        path: (str) the path to read
        chunksize: (int) the chunksize, in samples
        start: (float) time to skip before reading
        stop: (float) time to stop reading (0=end of file)

    Returns:
        a generator yielding numpy arrays (float64) of at most `chunksize` frames

    Example
    ~~~~~~~

    .. code::

        >>> with sndwrite_chunked("out.flac", 44100) as writer:
        ...     for buf in sndread_chunked("in.flac"):
        ...         # do some processing, like changing the gain
        ...         buf *= 0.5
        ...         writer.write(buf)

    """
    path = _normalize_path(path)
    backend = _get_backend(path, key=lambda backend: backend.can_read_chunked)
    if backend:
        logger.debug(f"sndread_chunked: using backend {backend.name}")
        return backend.read_chunked(path, chunksize, start=start, stop=stop)
    else:
        raise SndfileError("chunked reading is not supported by the "
                           "available backends")


def sndinfo(path: str) -> SndInfo:
    """
    Get info about a soundfile. Returns a :class:`SndInfo`

    Args:
        path: (str) the path to a soundfile

    Returns:
        a :class:`SndInfo` (attributes: **samplerate**: `int`, *nframes*: `int`, *channels*: `int`,
        **encoding**: `str`, **fileformat**: `str`, **metadata**: `dict`)

    Example
    ~~~~~~~
    ::

        >>> from sndfileio import *
        >>> info = sndinfo("sndfile.wav")
        >>> print(f"Duration: {info.duration}s, samplerate: {info.samplerate}")
        Duration: 0.4s, samplerate: 44100

    """
    path = _normalize_path(path)
    backend = _get_backend(path)
    if not backend:
        raise FormatNotSupported("sndinfo: no backend supports this filetype")
    logger.debug(f"sndinfo: using backend {backend.name}")
    return backend.getinfo(path)


def sndget(path: str, start: float = 0, end: float = 0) -> Tuple[np.ndarray, SndInfo]:
    """
    Read a soundfile and its metadata

    Args:
        path: the path to read
        start: the time to start reading
        end: the time to end reading (0=until the end)

    Returns:
        a tuple (samples: np.ndarray, :class:`SndInfo`)

    Example
    ~~~~~~~
    ::

        # Normalize and save as flac, keeping the metadata
        from sndfileio import *
        samples, info = sndget("in.wav")
        maxvalue = max(samples.max(), -samples.min())
        samples *= 1/maxvalue
        sndwrite("out.flac", samples, info.samplerate, metadata=info.metadata)

    """
    path = _normalize_path(path)
    backend = _get_backend(path)
    if not backend:
        raise RuntimeError(f"No backend available to read {path}")
    logger.debug(f"sndread: using backend {backend.name}")
    return backend.read_with_info(path, start=start, end=end)


def _resolve_encoding(outfile: str, fileformat: Opt[str], encoding: Opt[str],
                      samples: Opt[np.ndarray] = None
                      ) -> Tuple[str, str]:
    if not fileformat:
        fileformat = util.fileformat_from_ext(_os.path.splitext(outfile)[1])
    if encoding == 'auto':
        if samples:
            encoding = util.guess_encoding(samples, fileformat)
        else:
            encoding = util.default_encoding(fileformat)
    elif encoding == 'default' or encoding is None:
        encoding = util.default_encoding(fileformat)
    return fileformat, encoding


def sndwrite(outfile: str, samples: np.ndarray, sr: int, encoding='default',
             fileformat: str = None, normalize_if_clipping=True,
             metadata: Dict[str, str] = None, **options
             ) -> None:
    """
    Write all samples to a soundfile.

    Args:
        outfile: The name of the outfile. the extension will determine
            the file-format. The formats supported depend on the available
            backends.
        samples: Array-like. the actual samples. The shape determines the
            number of channels of outfile. For 1 channel, ``shape=(nframes,)`` or
            ``shape=(nframes, 1)``. For multichannel audio, ``shape=(nframes, nchannels)``.
        sr: Sampling-rate
        encoding: one of "pcm16", "pcm24", "pcm32", "pcm64", "float32", "float64",
            "auto" to guess an encoding and "default" to use the default encoding
            for the given format.
        fileformat: if given, will override the format indicated by the extension
        normalize_if_clipping: prevent clipping by normalizing samples before
            writing. This only makes sense when writing pcm data
        metadata: a dict of str: str, with possible keys 'title', 'comment', 'artist',
            'album', 'tracknumber', 'software' (the creator of a soundfile)
        options: available options depend on the fileformat.

    .. admonition:: Options

        **mp3**:
            - `bitrate`: bitrate in Kb/s (int, default=160)
            - `quality`: 1-7, 1 is highest and 7 is fastest, default=3

        For lossless formats, like `wav`, `aif`, `flac`, etc., **there are no extra options**

    .. note::

        Not all file formats support all encodings. Raises :class:`SndfileError`
        if the format does not support the given encoding.
        If set to 'auto', an encoding will be selected based on the
        file-format and on the data. The bitdepth of the data is
        measured, and if the file-format supports it, it will be used.
        For bitdepths of 8, 16 and 24 bits, a PCM encoding will be used.
        For a bitdepth of 32 bits, a FLOAT encoding will be used,
        or the next lower supported encoding.
        If 'default' is given as encoding, the default encoding for the format
        is used

    =======    =================
    Format     Default encoding
    =======    =================
     wav        float32
     aif        float32
     flac       pcm24
     mp3        pcm16
    =======    =================


    Example
    ~~~~~~~
    ::

        # Normalize and save as flac
        >>> samples, sr = sndread("sndfile.wav")
        >>> maxvalue = max(samples.max(), -samples.min())
        >>> samples *= 1/maxvalue
        >>> sndwrite("out.flac", samples, sr)

    """
    outfile = _normalize_path(outfile)
    fileformat, encoding = _resolve_encoding(outfile, samples=samples, fileformat=fileformat,
                                             encoding=encoding)
    if encoding.startswith('pcm') and normalize_if_clipping:
        clipping = ((samples > 1).any() or (samples < -1).any())
        if clipping:
            maxvalue = max(samples.max(), abs(samples.min()))
            logger.warning(f"Clipping found when writing pcm data to {outfile}")
            samples = samples / maxvalue
    backend = _get_write_backend(fileformat)
    if not backend:
        raise SndfileError(f"No backend found to support the given format: {fileformat}")
    logger.debug(f"sndwrite: using backend {backend.name}")
    writer = backend.writer(sr=sr, outfile=outfile, encoding=encoding, metadata=metadata,
                            fileformat=fileformat, **options)
    if not writer:
        raise SndfileError(f"Could not write to {outfile} with backend {backend.name}")
    return writer.write(samples)


def sndwrite_chunked(outfile: str, sr: int, encoding='auto', fileformat: str = None,
                     metadata: Dict[str, str] = None, **options
                     ) -> SndWriter:
    """
    Opens a file for writing and returns a SndWriter

    The :meth:`~SndWriter.write` method of the returned :class:`SndWriter` can be
    called to write samples to the file

    Raises SndfileError if the format does not support the given encoding.

    Not all file formats support all encodings. If set to 'auto', an encoding will
    be selected based on the file-format and on the data. The bitdepth of the data is
    measured, and if the file-format supports it, it will be used.
    For bitdepths of 8, 16 and 24 bits, a PCM encoding will be used.
    For a bitdepth of 32 bits, a FLOAT encoding will be used,
    or the next lower supported encoding

    Args:
        outfile: The name of the outfile. the extension will determine the file-format.
            The formats supported depend on the available backends.
        sr: Sampling-rate
        encoding: one of 'auto', 'pcm16', 'pcm24', 'pcm32', 'float32', 'float64'.
        fileformat: needed only if the format cannot be determined from the extension
            (for example, if saving to an outfile with a non-traditional extension)
        metadata: a dict ``{str: str}`` with possible keys: 'comment', 'title', 'artist',
            'album', 'tracknumber', 'software' (the creator of a soundfile)
        options: available options depend on the fileformat.
            For mp3, options are: bitrate (int, default=128) and quality (1-7, where
            1 is highest and 7 is fastest, default=2)
            For non-destructive formats, like wav, aif, flac, etc., there are no extra options

    Returns:
        a :class:`~sndfileio.SndWriter`, whose method :meth:`~sndfileio.SndWriter.write` can
        be called to write samples

    Example
    ~~~~~~~

    ::

        from snfileio import *
        with sndwrite_chunked("out.flac", 44100) as writer:
            for buf in sndread_chunked("in.flac"):
                # do some processing, like changing the gain
                buf *= 0.5
                writer.write(buf)

    """
    outfile = _normalize_path(outfile)
    fileformat, encoding = _resolve_encoding(outfile, fileformat=fileformat,
                                             encoding=encoding)
    backends = [backend for backend in _get_backends()
                if backend.can_write_chunked and fileformat in backend.filetypes_write]
    if not backends:
        raise SndfileError(f"No backend found to support the format {fileformat}")
    backend = min(backends, key=lambda backend: backend.priority)
    logger.debug(f"sndwrite_chunked: using backend {backend.name}")
    return backend.writer(outfile, sr, encoding, metadata=metadata, fileformat=fileformat,
                          **options)


def sndwrite_like(outfile: str, samples: np.ndarray, likefile: str, sr: int = None,
                  metadata: Dict[str, str] = None
                  ) -> None:
    """
    Write samples to outfile with samplerate/fileformat/encoding taken from likefile

    Args:
        outfile: the file to write to
        samples: the samples to write
        likefile: the file to use as a reference for sr, format and encoding
        sr: sample rate can be overridden
        metadata: a dict {str: str}, overrides metadata in `likefile`. Metadata
            is not merged, so if metadata is given, it substitutes the metadata
            in `likefile` completely. In order to merge it, do that beforehand
            If None is passed, the metadata in `likefile` is written to `outfile`

    .. note::

        The fileformat is always determined by `likefile`, even if the extension
        of `outfile` would result in a different format. For example, if `likefile`
        has a flac format but outfile has a .wav extension, the resulting file will
        be written in flac format.

    Example
    ~~~~~~~

    ::

        # Read a file, apply a fade-in of 0.5 seconds, save it
        import numpy as np
        from sndfileio import *
        samples, sr = sndread("stereo.wav")
        fadesize = int(0.5*sr)
        ramp = np.linspace(0, 1, fadesize))
        samples[:fadesize, 0] *= ramp
        samples[:fadesize, 1] *= ramp
        sndwrite_like(samples, "stereo.wav", "out.wav")

    """
    outfile = _normalize_path(outfile)
    info = sndinfo(likefile)
    ext = _os.path.splitext(outfile)[1]
    outfileformat = util.fileformat_from_ext(ext)
    if outfileformat != info.fileformat:
        logger.warning(f"Trying to save to a file with extension {ext}, but fileformat"
                       f"will be {info.fileformat}")
    sndwrite(outfile=outfile, samples=samples, sr=sr or info.samplerate,
             fileformat=info.fileformat, encoding=info.encoding,
             metadata=metadata or info.metadata)


def sndwrite_chunked_like(outfile: str, likefile: str, sr: int = None,
                          metadata: Dict[str, str] = None
                          ) -> SndWriter:
    """
    Create a SndWriter with samplerate/format/encoding of the
    source file

    Args:
        outfile: the file to open for writing
        likefile: the file to use as reference
        sr: samplerate can be overridden
        metadata: a dict {str: str}, overrides metadata in `likefile`. Metadata
            is not merged, so if metadata is given, it substitutes the metadata
            in `likefile` completely. In order to merge it, do that beforehand
            If None is passed, the metadata in `likefile` is written to `outfile`

    Returns:
        a :class:`SndWriter`. Call :meth:`~SndWriter.write` on it to write to
        the file
    """
    info = sndinfo(likefile)
    return sndwrite_chunked(outfile=outfile, sr=sr or info.samplerate,
                            encoding=info.encoding, metadata=metadata or info.metadata)


def mp3write(outfile: str, samples: np.ndarray, sr: int, bitrate=224, quality=3
             ) -> None:
    """
    Write all samples to outfile as mp3

    This is the same as::

        sndwrite(outfile, samples, sr, bitrate=224, quality=3)

    But in this case the arguments are explictely listed instead of being part of
    ``**options``

    Args:
        outfile: the outfile to write to
        samples: the samples to write (float samples in the range -1:1). They will
            be converted to int16 so any values outside the given range will clip
        sr: the samplerate
        bitrate: the bitrate to use, in Kb/s
        quality: the quality, a value between 1-7 (where 1 is highest and 7 is fastest)

    .. note::
        To write samples in chunks use sndwrite_chunked. `bitrate` and `quality` can
        be passed as **options.

    """
    outfile = _normalize_path(outfile)
    mp3backend = _BACKENDS['lameenc']
    if not mp3backend.is_available():
        raise RuntimeError("lameenc backend is not available")
    writer = mp3backend.writer(outfile=outfile, sr=sr, encoding='pcm16', fileformat='mp3',
                               bitrate=bitrate, quality=quality)
    writer.write(samples)


############################################
#
#                BACKENDS
#
############################################

def _asbytes(s: Union[str, bytes]) -> bytes:
    if isinstance(s, bytes):
        return s
    return s.encode('ascii')


class _PySndfileWriter(SndWriter):
    _keyTable = {
        'comment': 'SF_STR_COMMENT',
        'title': 'SF_STR_TITLE',
        'artist': 'SF_STR_ARTIST',
        'album': 'SF_STR_ALBUM',
        'tracknumber': 'SF_STR_TRACKNUMBER',
        'software': 'SF_STR_SOFTWARE'
    }

    def _open_file(self, channels: int) -> None:
        if self.fileformat not in self.filetypes:
            raise ValueError(f"Format {self.fileformat} not supported by this backend")
        sndformat = self._backend.get_sndfile_format(self.fileformat, self.encoding)
        self._file = self._backend.PySndfile(self.outfile, "w", sndformat, channels, self.sr)
        if self.metadata:
            for k, v in self.metadata.items():
                key = self._keyTable[k]
                self._file.set_string(key, _asbytes(v))

    def write(self, frames: np.ndarray) -> None:
        if not self._file:
            nchannels = util.numchannels(frames)
            if self.encoding == 'auto':
                self.encoding = util.guess_encoding(frames, self.fileformat)
            elif self.encoding == 'default':
                self.encoding = util.default_encoding(self.fileformat)
            self._open_file(nchannels)
        self._file.write_frames(frames)

    def close(self):
        if self._file is None:
            raise IOError("Can't close, since this file was never open")
        self._file.writeSync()
        del self._file
        self._file = None


class Backend:
    def __init__(self, priority: int,
                 filetypes: List[str],
                 filetypes_write: List[str],
                 can_read_chunked: bool,
                 can_write_chunked: bool,
                 name: str,
                 supports_metadata: bool):
        self.priority = priority
        self.filetypes = filetypes
        self.filetypes_write = filetypes_write
        self.can_read_chunked = can_read_chunked
        self.can_write_chunked = can_write_chunked
        self.supports_metadata = supports_metadata
        self.name = name
        self._backend = None
        self._writer: Opt[Type[SndWriter]] = None

    def read_with_info(self, path: str, start=0., end=0.
                       ) -> Tuple[np.ndarray, SndInfo]:
        return NotImplemented

    def read(self, path: str, start=0., end=0.) -> sample_t:
        samples, info = self.read_with_info(path=path, start=start, end=end)
        return samples, info.samplerate

    def read_chunked(self, path: str, chunksize=2048, start=0., stop=0.
                     ) -> Iterator[np.ndarray]:
        return NotImplemented

    def getinfo(self, path: str) -> SndInfo:
        """ Get info about the soundfile given. Returns a SndInfo structure """
        return NotImplemented

    def is_available(self) -> bool:
        """ Is this backend available? """
        return _is_package_installed(self.name)

    def writer(self, outfile: str, sr: int, encoding: str, fileformat: str,
               bitrate=0, metadata: Dict[str, str] = None, **options
               ) -> SndWriter:
        """ Open outfile for write with the given properties

        Args:
            sr: samplerate
            outfile: the file to write
            encoding: the encoding used (pcm16, float32, etc)
            fileformat: the file format
            bitrate: used when writing to a compressed file.
            metadata: if given, a dict str: str. Allowed keys are: *title*, *comment*,
                *artist*, *tracknumber*, *album*, *software*

        Returns:
            a :class:`SndWriter`
        """
        pass
        if self._writer is None:
            raise SndfileError(f"The backend {self.name} does not support writing")
        return self._writer(backend=self, sr=sr, outfile=outfile, encoding=encoding,
                            fileformat=fileformat, bitrate=bitrate, metadata=metadata,
                            **options)

    def check_write(self, fileformat: str, encoding: str) -> None:
        """ Check if we can write to outfile with the given encoding """
        if encoding not in util.encodings_for_format[fileformat]:
            raise ValueError("Encoding not supported")
        if fileformat not in self.filetypes_write:
            raise ValueError(f"The given format {fileformat} is not supported by the "
                             f"{self.name} backend")

    def dump(self) -> None:
        """ Dump information about this backend """
        print(f"Backend: {self.name} (available: {self.is_available()}, priority: {self.priority})")
        if self.filetypes:
            readtypes = ", ".join(self.filetypes)
            print(f"    read types : {readtypes}")
        if self.filetypes_write:
            writetypes = ", ".join(self.filetypes_write)
            print(f"    write types: {writetypes}")
        ok, notok = "OK", "--"
        readchunked = ok if self.can_read_chunked else notok
        writechunked = ok if self.can_write_chunked else notok
        print(f"    sndread_chunked: {readchunked}    sndwrite_chunked: {writechunked}")


class _Lameenc(Backend):

    def __init__(self, priority: int):
        super().__init__(priority=priority,
                         filetypes=[],
                         filetypes_write=['mp3'],
                         can_read_chunked=False,
                         can_write_chunked=True,
                         name='lameenc',
                         supports_metadata=False)

    def writer(self, outfile: str, sr: int, encoding: str, fileformat: str,
               metadata: Dict[str, str] = None, **options
               ) -> SndWriter:
        from . import backend_lameenc
        bitrate = options.pop('bitrate', 160)
        quality = options.pop('quality', 3)
        return backend_lameenc.LameencWriter(outfile=outfile, sr=sr, bitrate=bitrate,
                                             quality=quality)


class _SoundfileWriter(SndWriter):
    def _open_file(self, channels: int) -> None:
        if self.fileformat not in self.filetypes:
            raise ValueError(f"Format {self.fileformat} not supported by this backend")
        fmt, subtype = _Soundfile.get_format_and_subtype(self.fileformat, self.encoding)
        self._file = _soundfile.SoundFile(self.outfile, "w", format=fmt, subtype=subtype,
                                          channels=channels, samplerate=self.sr)
        if self.metadata:
            for key, value in self.metadata.items():
                setattr(self._file, key, value)

    def write(self, frames: np.ndarray) -> None:
        if not self._file:
            nchannels = util.numchannels(frames)
            if self.encoding == 'auto':
                self.encoding = util.guess_encoding(frames, self.fileformat)
            elif self.encoding == 'default':
                self.encoding = util.default_encoding(self.fileformat)
            self._open_file(nchannels)
        self._file.write(frames)

    def close(self):
        if self._file is None:
            return
        self._file.close()
        self._file = None


class _Soundfile(Backend):
    """
    A backend based on soundfile (https://pysoundfile.readthedocs.io)

    """
    # TODO: support metadata (either PR or via
    #  https://github.com/thebigmunch/audio-metadata)
    subtype_to_encoding = {
        'PCM_24': 'pcm24',
        'PCM_16': 'pcm16',
        'PCM_32': 'pcm32',
        'PCM_64': 'pcm64',
        'FLOAT': 'float32',
        'DOUBLE': 'float64',
        'VORBIS': 'vorbis'
    }

    encoding_to_subtype = {
        'pcm16': 'PCM_16',
        'pcm24': 'PCM_24',
        'float32': 'FLOAT',
        'float64': 'DOUBLE',
    }

    format_to_fileformat = {
        'WAV': 'wav',
        'WAVEX': 'wav',
        'AIFF': 'aiff',
        'FLAC': 'flac',
        'OGG': 'ogg'
    }

    # 'comment', 'title', 'artist',
    #             'album', 'tracknumber', 'software'
    metadata_keys = {'comment', 'title', 'artist', 'album', 'tracknumber', 'software'}

    def __init__(self, priority: int):
        super().__init__(
            priority=priority,
            filetypes=["aif", "aiff", "wav", "flac", "ogg"],
            filetypes_write=["aif", "aiff", "wav", "flac"],
            can_read_chunked=True,
            can_write_chunked=True,
            name='soundfile',
            supports_metadata=True
        )
        self._writer = _SoundfileWriter

    def read_with_info(self, path: str, start: float = 0, end: float = 0) -> Tuple[np.ndarray, SndInfo]:
        snd = _soundfile.SoundFile(path, 'r')
        info = self._getinfo(snd)
        samples = self._read(snd, start=start, end=end)
        return samples, info

    def _read(self, snd: _soundfile.SoundFile, start: float = 0, end: float = 0) -> np.ndarray:
        sr: int = snd.samplerate
        samp0 = int(start*sr)
        samp1 = int(end*sr) if end > 0 else snd.frames
        if samp0:
            snd.seek(samp0)
        return snd.read(samp1 - samp0)

    def read_chunked(self, path: str, chunksize=2048, start=0., stop=0.
                     ) -> Iterator[np.ndarray]:
        snd = _soundfile.SoundFile(path, 'r')
        sr = snd.samplerate
        if start:
            snd.seek(int(start*snd.samplerate))
        firstframe = int(sr * start)
        lastframe = snd.frames if stop == 0 else int(sr*stop)
        for pos, nframes in util.chunks(0, lastframe - firstframe, chunksize):
            yield snd.read(nframes)

    def getinfo(self, path: str) -> SndInfo:
        return self._getinfo(_soundfile.SoundFile(path, 'r'))

    def _getinfo(self, snd: _soundfile.SoundFile, ) -> SndInfo:
        encoding = _Soundfile.subtype_to_encoding[snd.subtype]
        fileformat = _Soundfile.format_to_fileformat[snd.format]
        metadata = {}
        for key in self.metadata_keys:
            value = getattr(snd, key, None)
            if value:
                metadata[key] = value

        ext = _os.path.splitext(snd.name)[1]
        if ext == '.ogg':
            metadata = util.tinytagMetadata(snd.name)
            bitrate = metadata.pop('bitrate', None)
        else:
            bitrate = None

        return SndInfo(samplerate=snd.samplerate,
                       nframes=snd.frames,
                       channels=snd.channels,
                       encoding=encoding,
                       fileformat=fileformat,
                       metadata=metadata,
                       bitrate=bitrate)

    @staticmethod
    def get_format_and_subtype(fmt: str, encoding: str = None
                               ) -> Tuple[str, str]:
        soundfile_fmt = {
            'wav': 'WAVEX',
            'aif': 'AIFF',
            'aiff': 'AIFF',
            'flac': 'FLAC',
        }.get(fmt)
        if not soundfile_fmt:
            raise ValueError(f"Format {fmt} not supported")
        subformat = _Soundfile.encoding_to_subtype.get(encoding)
        if not subformat:
            raise ValueError(f"Encoding {encoding} not supported for {fmt}")
        return soundfile_fmt, subformat


class _PySndfile(Backend):
    """
    A backend based in pysndfile

    """
    _keyTable = {'SF_STR_COMMENT': 'comment',
                 'SF_STR_TITLE': 'title',
                 'SF_STR_ARTIST': 'artist',
                 'SF_STR_ALBUM': 'album',
                 'SF_STR_TRACKNUMBER': 'tracknumber',
                 'SF_STR_SOFTWARE': 'software'}
    # _writer = _PySndfileWriter

    def __init__(self, priority: int):

        super().__init__(
                priority  = priority,
                filetypes = ["aif", "aiff",  "wav", "flac", "ogg", "wav64", "caf", "raw"],
                filetypes_write = ["aif", "aiff",  "wav", "flac", "ogg", "wav64", "caf", "raw"],
                can_read_chunked = True,
                can_write_chunked = True,
                name = 'pysndfile',
                supports_metadata= True
        )
        self._writer = _PySndfileWriter
        import pysndfile
        self.pysndfile = pysndfile

    def read_with_info(self, path: str, start: float = 0, end: float = 0) -> Tuple[np.ndarray, SndInfo]:
        snd = self.pysndfile.PySndfile(path)
        info = self._getinfo(snd)
        samples = self._read(snd, start=start, end=end)
        return samples, info

    def _read(self, snd: pysndfile.PySndfile, start=0., end=0.) -> np.ndarray:
        sr: int = snd.samplerate()
        samp_start = int(start * sr)
        samp_end = int(end * sr) if end > 0 else snd.frames()
        if samp_start:
            snd.seek(samp_start)
        return snd.read_frames(samp_end - samp_start)

    def read_chunked(self, path: str, chunksize=2048, start: float = 0., stop: float = 0.
                     ) -> Iterator[np.ndarray]:
        snd = self.pysndfile.PySndfile(path)
        sr = snd.samplerate()
        if start:
            snd.seek(int(start*snd.samplerate()))
        firstframe = int(sr * start)
        lastframe = snd.frames() if stop == 0 else int(sr*stop)
        for pos, nframes in util.chunks(0, lastframe - firstframe, chunksize):
            yield snd.read_frames(nframes)

    def getinfo(self, path: str) -> SndInfo:
        return self._getinfo(self.pysndfile.PySndfile(path))

    def _getinfo(self, snd: pysndfile.PySndfile) -> SndInfo:
        metadataraw: Dict[str, bytes] = snd.get_strings()
        if metadataraw:
            metadata, extrainfo = {}, {}
            for k, v in metadataraw.items():
                if k not in self._keyTable:
                    extrainfo[k] = v
                else:
                    metadata[self._keyTable[k]] = v
        else:
            metadata, extrainfo = None, None

        return SndInfo(snd.samplerate(), snd.frames(), snd.channels(),
                       snd.encoding_str(), snd.major_format_str(),
                       metadata=metadata, extrainfo=extrainfo)

    def get_sndfile_format(self, fileformat: str, encoding: str) -> int:
        """
        Construct a pysndfile format id from fileformat and encoding

        Args:
            fileformat: the fileformat, one of 'wav', 'aiff', etc
            encoding: one of *pcmXX* or *floatXX* (where *XX *is the number of bits/sample,
                one of 16, 24, 32, 64)

        Returns:
            the pysndfile format id
        """
        assert fileformat in self.filetypes
        fmt, bits = encoding[:-2], int(encoding[-2:])
        assert fmt in ('pcm', 'float') and bits in (8, 16, 24, 32, 64)
        if fileformat == 'aif':
            fileformat = 'aiff'
        return self.pysndfile.construct_format(fileformat, f"{fmt}{bits}")

    def detect_format(self, path: str) -> Opt[str]:
        f = self.pysndfile.PySndfile(path)
        return self.pysndfile.fileformat_id_to_name.get(f.format())


# ----------------------------------------------------------------

class _Miniaudio(Backend):

    def __init__(self, priority):
        super().__init__(
                priority=priority,
                filetypes= ['mp3', 'ogg'],
                filetypes_write = [],
                can_read_chunked = True,
                can_write_chunked = False,
                name = 'miniaudio',
                supports_metadata=False
        )

    def getinfo(self, path: str) -> SndInfo:
        from . import backend_miniaudio
        ext = _os.path.splitext(path)[1]
        if ext == '.mp3':
            return backend_miniaudio.mp3info(path)
        elif ext == '.ogg':
            return backend_miniaudio.ogginfo(path)

    def read_with_info(self, path: str, start=0., end=0.) -> Tuple[np.ndarray, SndInfo]:
        return self.read(path, start, end)[0], self.getinfo(path)

    def read(self, path: str, start: float = 0., end: float = 0.) -> sample_t:
        from . import backend_miniaudio
        return backend_miniaudio.mp3read(path, start=start, end=end)

    def read_chunked(self, path: str, chunksize=2048, start: float = 0., stop: float = 0.
                     ) -> Iterator[np.ndarray]:
        from . import backend_miniaudio
        ext = _os.path.splitext(path)[1]
        if ext == '.mp3':
            return backend_miniaudio.mp3read_chunked(path, chunksize=chunksize, start=start, stop=stop)
        else:
            raise FormatNotSupported(f'chunked reading is not supported for {ext}')


_BACKENDS: Dict[str, Backend] = {
    'soundfile': _Soundfile(priority=0),
    'lameenc': _Lameenc(priority=10),
    'miniaudio': _Miniaudio(priority=10),
}

# if _is_package_installed('pysndfile'):
#     _BACKENDS['pysndfile'] = _PySndfile(priority=1)


_cache = {}


def report_backends():
    for b in _BACKENDS.values():
        if b.is_available():
            b.dump()
        else:
            print(f"Backend {b.name} NOT available")


def _get_backends() -> List[Backend]:
    backends = _cache.get('backends')
    if not backends:
        _cache['backends'] = backends = [b for b in _BACKENDS.values() if b.is_available()]
    return backends


def _get_backend(path: str = None, key: Callable[[Backend], bool] = None
                 ) -> Opt[Backend]:
    """
    Get available backends to read/write the file given

    Args:
        path: the file to read/write
        key: a function (backend) -> bool signaling if the backend
             is suitable for a specific task

    Example
    ~~~~~~~
    ::

        # Get available backends which can read in chunks
        >>> backend = _get_backend('file.flac',
        ...                        key=lambda backend:backend.can_read_chunked())
    """
    filetype = util.detect_format(path)
    backends = _get_backends()
    if key:
        backends = [b for b in backends if key(b)]
    if filetype:
        backends = [b for b in backends if filetype in b.filetypes]
    if backends:
        return min(backends, key=lambda backend: backend.priority)
    return None


def _get_write_backend(fileformat: str) -> Opt[Backend]:
    backends = _get_backends()
    if not backends:
        raise SndfileError("No available backends for writing")
    backends = [b for b in backends if fileformat in b.filetypes_write]
    if backends:
        return min(backends, key=lambda backend: backend.priority)
    return None
